/**************************************************************************/
/*                                Flow4J                                  */
/*             A visual process flow configuration framework              */
/*                         (c) Alex Greif 2003                            */
/*                                                                        */
/*  This program is free software; you can redistribute it and/or modify  */
/*  it under the terms of the GNU General Public License as published by  */
/*  the Free Software Foundation; either version 2 of the License, or     */
/*  (at your option) any later version.                                   */
/*                                                                        */
/**************************************************************************/

package flow4j.flow;

import java.io.*;
import java.util.*;

import flow4j.flow.ITask;


/**	The FolwMgr is a singleton that is able to register folws in a
 specified folder, create flowlets and to start flow execution.
 */
public class FlowMgr {


	/**	The sole instance of this singleton FlowMgr.
	 */
	static private FlowMgr _flowMgr;

	/**	Container for registered flows. Flows are stored as singletons
	 Key: flowName
	 Value: the Flow
	 */
	static private Map _flows = new HashMap();





	/** Private constructor
	 */
	private FlowMgr() {
	}


	/** Returns the single instance of the flow manager.
	 The instance will be created if not yet existent
	 */
	static public FlowMgr getFlowMgr() {
		if (_flowMgr == null)
			_flowMgr = new FlowMgr();

		return _flowMgr;
	}




	/**	Creates and returns a new instance of the flow and registeres it in the flows Map.
	 */
	static private IFlow registerFlow(String flowName) {
		IFlow flow = null;
		String fullQualifFlowName = getFlowClassName(flowName);
		try
		{
			//	create new instance and put in map
			flow = (IFlow)Class.forName(fullQualifFlowName).newInstance();
			_flows.put(flowName, flow);
		} catch (Exception e) {
			throw new FlowException("Could not register flow: " + fullQualifFlowName, e);
		}

		return flow;
	}


	/**	Returns the class name of the flow. Every flow resides by default in the
	 flow package
	 */
	private static String getFlowClassName(String flowName) {
		return "flows." + flowName.trim();
	}



	/**	Starts the flow with the given enty point as String.
	 The delimiter for flow and entry is a hyphen.
	 */
	static public void start(String flowEntryString) {
		start(flowEntryString, new FlowDictionary());
	}


	/**	Starts the flow with the given enty point as String.
	 The delimiter for flow and entry is a hyphen.
	 */
	static public void start(String flowEntryString, FlowDictionary dictionary) {
		int hyphenPos = flowEntryString.indexOf("-");
		String flowName = flowEntryString.substring(0, hyphenPos);
		String flowEntry = flowEntryString.substring(hyphenPos+1);

		start(flowName, flowEntry, dictionary);
	}


	/**	Starts the flow with the given enty point.
	 */
	static public void start(String flowName, String flowEntry) {
		start(flowName, flowEntry, new FlowDictionary());
	}


	/**	Starts the flow with the given enty point and dictionary.
	 If the flow is not registered yet, then a new instance is registered,
	 otherwise the singleton instance of the flow is returned
	 */
	static public void start(String flowName, String flowEntry, FlowDictionary dictionary) {
		IFlow flow = (IFlow)_flows.get(flowName.trim());
		if (flow == null)
			flow = registerFlow(flowName);

		flow.start(flowEntry, dictionary);
	}





	/**	Returns a new instance of a task
	 @param fullQualifClassName the fully qualified class name of the task class.
	 */
	static public ITask getTaskInstance(String fullQualifClassName) {
		try {
			Class clazz = Class.forName(fullQualifClassName);
			ITask task = (ITask)clazz.newInstance();
			return task;
		} catch (ClassNotFoundException e) {
			throw new FlowException("Task class not found: "
										+ fullQualifClassName, e);

		} catch (InstantiationException e) {
			throw new FlowException("Could not create instance of Task: "
										+ fullQualifClassName, e);
		} catch (IllegalAccessException e) {
			throw new FlowException("Could not create instance of Task: "
										+ fullQualifClassName, e);
		}
	}


	/*
	static public ITask getTask(Class clazz) {
		try {
			ITask flowlet = (ITask)clazz.newInstance();
			return flowlet;
		} catch (Exception e) {
			throw new FlowException("Could not create instance of flowlet: "
										+ clazz.getName(), e);
		}
	}
	 */



}

