/**************************************************************************/
/*                                Flow4J                                  */
/*             A visual process flow configuration framework              */
/*                         (c) Alex Greif 2003                            */
/*                                                                        */
/*  This program is free software; you can redistribute it and/or modify  */
/*  it under the terms of the GNU General Public License as published by  */
/*  the Free Software Foundation; either version 2 of the License, or     */
/*  (at your option) any later version.                                   */
/*                                                                        */
/**************************************************************************/

package flow4j.kopi;

import java.io.*;
import java.util.*;
import java.text.MessageFormat;

import at.dms.compiler.*;
import at.dms.kjc.*;
import at.dms.compiler.tools.antlr.extra.InputBuffer;
import at.dms.compiler.tools.antlr.runtime.ParserException;
import at.dms.util.Message;
import at.dms.util.MessageDescription;
import at.dms.classfile.ClassFileFormatException;




/**
 */
public class IMCompiler extends at.dms.kjc.Main {


	private int version = 0;
	// all generated classes
	private Vector classes = new Vector(100);

	private List troubles = new ArrayList();


	/**
	 *
	 */
	public static void main(String[] args)
	{
		String source = "public class Bla{Object o1 = \"a\"; \nObject o2 = 1;}";
		IMCompiler compiler = new IMCompiler();
		compiler.compile(source);
	}

	/**
	 */
	public JCompilationUnit parse(String source) {
		KjcEnvironment environment = createEnvironment(new KjcOptions());
		return parse(source, environment);
	}

	/**
	 *
	 */
	public JCompilationUnit parse(String source, KjcEnvironment environment) {
		setSourceVersion(environment.getSourceVersion());

		initialize(environment);

		JCompilationUnit unit = parseSource(source, environment);
		return unit;
	}



	/**
	 */
	public byte[] compile(String source) {
		KjcEnvironment environment = createEnvironment(new KjcOptions());
		JCompilationUnit unit = parse(source, environment);

		if (unit == null)
			return null;

		join(unit);
		checkInterface(unit);

		try {
			unit.checkInitializers(this, classes);
			unit.checkBody(this, classes);
		} catch (PositionedError e) {
			reportTrouble(e);
		}

		byte[] bytecode = generateCode(environment.getTypeFactory());
		return bytecode;
	}



	/**
	 * Creates a new compiler instance.
	 *
	 * @param	workingDirectory	the working directory
	 * @param	diagnosticOutput	the diagnostic output stream
	 */
	public IMCompiler(String workingDirectory, PrintWriter diagnosticOutput) {
		super(workingDirectory, diagnosticOutput);

	}

	public IMCompiler() {
		this(null, null);
	}


	/**
	 */
	public List getTroubles() {
		return troubles;
	}


	/**
	 * Runs a compilation session
	 *
	 * @param	args		the command line arguments
	 */
	public boolean run(String[] args) {
		return false;
	}




	/**
	 */
	public JCompilationUnit parseSource(String source, KjcEnvironment environment) {
		InputBuffer	buffer = null;

		Reader r = new StringReader(source);
		try
		{
			buffer = new InputBuffer("", r);
		}
		catch (IOException e) {
			e.printStackTrace();
		}

		KjcParser parser;
		JCompilationUnit unit;
		long lastTime = System.currentTimeMillis();

		parser = new KjcParser(this, buffer, environment);

		try {
			unit = parser.jCompilationUnit();
		} catch (ParserException e) {
			reportTrouble(parser.beautifyParseError(e));
			unit = null;
		} catch (Exception e) {
			//err.println("{" + file.getPath() + ":" + scanner.getLine() + "} " + e.getMessage());
			e.printStackTrace();
			//errorFound = true;
			unit = null;
		}

		try {
			buffer.close();
		} catch (IOException e) {
			e.printStackTrace();
		}

		return unit;

	}







	/**
	 * Generates the code from an array of compilation unit and
	 * a destination
	 *
	 * @param	destination	the directory where to write classfiles
	 */
	public byte[] generateCode(TypeFactory factory) {
		CSourceClass[]	classes = getClasses();
		BytecodeOptimizer	optimizer = new BytecodeOptimizer(1);

		this.classes.setSize(0);

		try {
			for (int count = 0; count < classes.length; count++) {
				System.out.println("" + classes[count]);

				ByteArrayOutputStream bout = new ByteArrayOutputStream();
				DataOutputStream outputStream =
					new DataOutputStream(bout);

				classes[count].genCode(optimizer, outputStream, factory);

				classes[count] = null;
				outputStream.flush();
				outputStream.close();

				return bout.toByteArray();
			}
		} catch (PositionedError e) {
			reportTrouble(e);
		} catch (ClassFileFormatException e) {
			e.printStackTrace();
			reportTrouble(new UnpositionedError(CompilerMessages.FORMATTED_ERROR, e.getMessage()));
		} catch (IOException e) {
			reportTrouble(new UnpositionedError(CompilerMessages.IO_EXCEPTION,
												"classfile",	//!!!FIXME !!!
												e.getMessage()));
		}

		return null;
	}











	/**
	 * Reports a trouble (error or warning).
	 *
	 * @param	trouble		a description of the trouble to report.
	 */
	public void reportTrouble(PositionedError trouble) {

		Message m = trouble.getFormattedMessage();
		MessageDescription md = m.getDescription();
		String text = format(md, m.getParams());
		//System.out.println("IMCompiler trouble: " + md.getLevel() + "   " + text
		//					   + "  line: " + trouble.getTokenReference().getLine());


		troubles.add(trouble);
	}



	/**
	 * Returns a string explaining the error.
	 *
	 * @param	parameters		the array of parameters
	 */
	public String format(MessageDescription md, Object[] parameters) {
		String	prefix;			// the text for the severity level
		String	body;			// the formatted message
		String	suffix;			// the reference

		switch (md.getLevel()) {
			case MessageDescription.LVL_UNDEFINED:
				// no qualifier
				prefix = "";
				break;
			case MessageDescription.LVL_ERROR:
				prefix = "error:";
				break;
			case MessageDescription.LVL_CAUTION:
				prefix = "caution:";
				break;
			case MessageDescription.LVL_WARNING:
				prefix = "warning:";
				break;
			case MessageDescription.LVL_NOTICE:
				prefix = "notice:";
				break;
			case MessageDescription.LVL_INFO:
				prefix = "";
				break;
			default:
				// unknown: mark as error
				prefix = "error:";
		}

		try {
			body = MessageFormat.format(md.getFormat(), parameters);
		} catch (RuntimeException e) {
			// wrong number of parameters: give at least message text with placeholders
			body = md.getFormat();
		}

		suffix = md.getReference() == null ? "" : " [" + md.getReference() + "]";

		return body;
	}




	/**
	 * Returns true iff comments should be parsed (false if to be skipped).
	 */
	public boolean parseComments() {
		return false;
	}

	/**
	 * Returns true iff compilation runs in verbose mode.
	 */
	public boolean verboseMode() {
		return false;
	}



	/**
	 * Returns the classes to generate
	 */
	public CSourceClass[] getClasses() {
		return (CSourceClass[])at.dms.util.Utils.toArray(classes, CSourceClass.class);
	}






	/**
	 */
	public Class loadClass(byte[] bytecode) throws ClassFormatError {
		return new CL(this.getClass().getClassLoader()).loadClass(bytecode);
	}






	/**
	 * A classloader for the compiled class
	 */
	private class CL extends ClassLoader {

		/** instantiated classes */
		private Map classInstances = new HashMap();

		/**
		 * Constructor
		 */
		private CL(ClassLoader parent) {
			super(parent);
		}

		/**
		 * @see java.lang.ClassLoader#findClass(java.lang.String)
		 */
		protected Class findClass(String key) throws ClassNotFoundException {
			// loaded it already?
			Class result = (Class)classInstances.get(key);
			if (result!=null) return result;
			// maybe it's an inner type?
			/*if (innerClasses!=null) {
				CompiledClass inner = (CompiledClass)innerClasses.get(key);
				if (inner!=null) return loadClass(inner);
			}
			 */
			// delegate
			return super.findClass(key);
		}

		/**
		 * Loads the class from the contained byte-data
		 * @return the Class
		 */
		private Class loadClass(byte[] bytecode) throws ClassFormatError {

			// Construct class
			Class result = defineClass(null, bytecode,0, bytecode.length);

			// Resolve it
			try {
				resolveClass(result);
			} catch (IncompatibleClassChangeError err) {
				// can't happen
			}

			// remember
			//classInstances.put(cc.name, result);

			// Done
			return result;
		}

	} //CL





}


